# Creates and populates two set[str]: all_files, all_dirs
def __belay_hfs(fns):
    buf = memoryview(bytearray(4096))
    return [__belay_hf(fn, buf) for fn in fns]
def __belay_mkdirs(fns):
    for fn in fns:
        try:
            os.mkdir(fn)
        except OSError:
            pass
def __belay_del_fs(path="/", keep=(), check=True):
    if not path:
        path = "/"
    elif not path.endswith("/"):
        path += "/"
    if check:
        try:
            os.stat(path)
        except OSError:
            return
    for name, mode, *_ in __belay_ilistdir(path):
        full_name = path + name
        if full_name in keep:
            continue
        if mode & 0x4000:  # is_dir
            __belay_del_fs(full_name, keep, check=False)
            try:
                os.rmdir(full_name)
            except OSError:
                pass
        else:
            os.remove(full_name)



import ast
from contextlib import nullcontext
from pathlib import Path
from typing import Dict, List, Optional, Set, Union
from urllib.parse import urlparse

import httpx
from rich.console import Console


class NonMatchingURL(Exception):
    pass


def _strip_www(url: str):
    if url.startswith("www."):
        url = url[4:]
    return url


def _process_url_github(url: str):
    """Transforms github-like url into githubusercontent."""
    url = str(url)
    parsed = urlparse(url)
    netloc = _strip_www(parsed.netloc)
    if netloc == "github.com":
        # Transform to raw.githubusercontent
        _, user, project, mode, branch, *path = parsed.path.split("/")
        return f"https://raw.githubusercontent.com/{user}/{project}/{branch}/{'/'.join(path)}"
    elif netloc == "raw.githubusercontent.com":
        return f"https://raw.githubusercontent.com{parsed.path}"
    else:
        # TODO: Try and be a little helpful if url contains github.com
        raise NonMatchingURL


def _process_url(url: str):
    parsers = [
        _process_url_github,
    ]
    for parser in parsers:
        try:
            return parser(url)
        except NonMatchingURL:
            pass

    # Unmodified URL
    return url


def _get_text(url: Union[str, Path]):
    url = str(url)
    if url.startswith(("https://", "http://")):
        res = httpx.get(url)
        res.raise_for_status()
        return res.text
    else:
        # Assume local file
        return Path(url).read_text()


def download_dependencies(
    dependencies: Dict[str, Union[str, Dict]],
    packages: Optional[List[str]] = None,
    local_dir: Union[str, Path] = ".belay/dependencies/main",
    console: Optional[Console] = None,
):
    """Download dependencies.

    Parameters
    ----------
    dependencies: dict
        Dependencies to install (probably parsed from TOML file).
    packages: Optional[List[str]]
        Only download this package.
    local_dir: Union[str, Path]
        Download dependencies to this directory.
        Will create directories as necessary.
    console: Optional[Console]
        Print progress out to console.
    """
    local_dir = Path(local_dir)
    if not packages:
        # Update all packages
        packages = list(dependencies.keys())

    if console:
        cm = console.status("[bold green]Updating Dependencies")
    else:
        cm = nullcontext()

    def log(*args, **kwargs):
        if console:
            console.log(*args, **kwargs)

    with cm:
        for pkg_name in packages:
            dep = dependencies[pkg_name]
            if isinstance(dep, str):
                dep = {"path": dep}
            elif not isinstance(dep, dict):
                raise ValueError(f"Invalid value for key {pkg_name}.")

            log(f"{pkg_name}: Updating...")

            url = _process_url(dep["path"])
            ext = Path(url).suffix

            # Single file
            dst = local_dir / (pkg_name + ext)
            dst.parent.mkdir(parents=True, exist_ok=True)

            new_code = _get_text(url)

            if ext == ".py":
                ast.parse(new_code)  # Check for valid python code

            try:
                old_code = dst.read_text()
            except FileNotFoundError:
                old_code = ""

            if new_code == old_code:
                log(f"{pkg_name}: No changes detected.")
            else:
                log(f"[bold green]{pkg_name}: Updated.")
                dst.write_text(new_code)


def clean_local(
    dependencies: Union[Set[str], List[str]],
    local_dir: Union[str, Path] = ".belay/dependencies/main",
):
    """Delete downloaded dependencies if they are no longer referenced."""
    local_dir = Path(local_dir)
    dependencies = set(dependencies)
    existing_deps = []
    existing_deps.extend(local_dir.glob("*.py"))

    for existing_dep in existing_deps:
        if existing_dep.stem in dependencies:
            continue
        existing_dep.unlink()



@micropython.native
def __belay_hf(fn, buf):
    # is inherently modulo 32-bit arithmetic
    @micropython.viper
    def xor_mm(data, state: uint, prime: uint) -> uint:
        for b in data:
            state = uint((state ^ uint(b)) * prime)
        return state

    h = 0x811c9dc5
    try:
        f = open(fn, "rb")
        while True:
            n = f.readinto(buf)
            if n == 0:
                break
            h = xor_mm(buf[:n], h, 0x01000193)
        f.close()
    except OSError:
        h = 0
    return h



from pathlib import Path

from typer import Argument, Option

from belay import Device
from belay.cli.common import help_password, help_port


def exec(
    port: str = Argument(..., help=help_port),
    statement: str = Argument(..., help="Statement to execute on-device."),
    password: str = Option("", help=help_password),
):
    """Execute python statement on-device."""
    device = Device(port, password=password)
    device(statement)
    device.close()



from belay.packagemanager import clean_local

from .common import load_toml


def clean():
    """Remove any downloaded dependencies if they are no longer specified in pyproject."""
    toml = load_toml()

    try:
        dependencies = toml["dependencies"]
    except KeyError:
        return

    clean_local(dependencies.keys())



# Don't manually change, let poetry-dynamic-versioning-plugin handle it.
__version__ = "0.14.2"

__all__ = [
    "AuthenticationError",
    "minify",
    "Device",
    "SpecialFunctionNameError",
    "PyboardException",
    "Implementation",
    "FeatureUnavailableError",
]
from ._minify import minify
from .device import Device, Implementation
from .exceptions import (
    AuthenticationError,
    ConnectionLost,
    FeatureUnavailableError,
    MaxHistoryLengthError,
    SpecialFunctionNameError,
)
from .pyboard import PyboardException



def __belay_hf(fn, buf):
    # inherently is inherently modulo 32-bit arithmetic
    mod = (1 << 32)
    h = 0x811c9dc5
    try:
        f = open(fn, "rb")
        while True:
            n = f.readinto(buf)
            if n == 0:
                break
            for b in buf[:n]:
                h = ((h ^ b) * 0x01000193) % mod  # todo: investigate fast mm
        f.close()
    except OSError:
        h = 0
    return h



import typer

from belay import Device
from belay.cli.clean import clean
from belay.cli.common import help_password, help_port
from belay.cli.exec import exec
from belay.cli.identify import identify
from belay.cli.info import info
from belay.cli.install import install
from belay.cli.new import new
from belay.cli.run import run
from belay.cli.sync import sync
from belay.cli.update import update

app = typer.Typer()
app.command()(clean)
app.command()(exec)
app.command()(identify)
app.command()(info)
app.command()(install)
app.command()(new)
app.command()(run)
app.command()(sync)
app.command()(update)



import inspect
from abc import abstractmethod
from functools import wraps
from typing import Callable, Optional

from autoregistry import Registry

from ._minify import minify as _minify
from .exceptions import FeatureUnavailableError, SpecialFunctionNameError
from .helpers import random_python_identifier, wraps_partial
from .inspect import getsource
from .typing import BelayCallable


class Executer(Registry, suffix="Executer"):
    def __init__(self, device):
        # To avoid Executer.__setattr__ raising an error
        object.__setattr__(self, "_belay_device", device)

    def __setattr__(self, name: str, value: BelayCallable):
        if (
            name.startswith("_belay")
            or name.startswith("__belay")
            or (name.startswith("__") and name.endswith("__"))
        ):
            raise SpecialFunctionNameError(
                f'Not allowed to register function named "{name}".'
            )
        super().__setattr__(name, value)

    def __getattr__(self, name: str) -> BelayCallable:
        # Just here for linting purposes.
        raise AttributeError

    @abstractmethod
    def __call__(self):
        raise NotImplementedError


class SetupExecuter(Executer):
    def __call__(
        self,
        f: Optional[BelayCallable] = None,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = True,
    ):
        """See ``Device.setup``."""
        if f is None:
            return wraps_partial(self, minify=minify, register=register, record=record)
        if inspect.isgeneratorfunction(f):
            raise ValueError("@Device.setup does not support generators.")
        name = f.__name__
        src_code, src_lineno, src_file = getsource(f, strip_signature=True)
        if minify:
            src_code = _minify(src_code)
        signature = inspect.signature(f)

        @wraps(f)
        def executer(*args, **kwargs):
            cmd = src_code
            bound_arguments = signature.bind(*args, **kwargs)
            bound_arguments.apply_defaults()
            arg_assign_cmd = "\n".join(
                f"{name}={repr(val)}" for name, val in bound_arguments.arguments.items()
            )
            if arg_assign_cmd:
                cmd = arg_assign_cmd + "\n" + cmd

            return self._belay_device._traceback_execute(
                src_file, src_lineno, name, cmd, record=record
            )

        if register:
            setattr(self, name, executer)

        return executer


class TaskExecuter(Executer):
    def __call__(
        self,
        f: Optional[BelayCallable] = None,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = False,
    ):
        """See ``Device.task``."""
        if f is None:
            return wraps_partial(self, minify=minify, register=register, record=record)

        name = f.__name__
        src_code, src_lineno, src_file = getsource(f)
        src_lineno -= 1  # Because of the injected ``@__belay`` decorator below

        # Add the __belay decorator for handling result serialization.
        src_code = f"@__belay({repr(name)})\n" + src_code

        # Send the source code over to the device.
        self._belay_device(src_code, minify=minify)

        @wraps(f)
        def func_executer(*args, **kwargs):
            cmd = f"_belay_{name}(*{repr(args)}, **{repr(kwargs)})"

            return self._belay_device._traceback_execute(
                src_file, src_lineno, name, cmd, record=record
            )

        @wraps(f)
        def gen_executer(*args, **kwargs):
            if record:
                raise NotImplementedError(
                    "Recording of generator tasks is currently not supported."
                )
            # Step 1: Create the on-device generator
            gen_identifier = random_python_identifier()
            cmd = f"{gen_identifier} = _belay_{name}(*{repr(args)}, **{repr(kwargs)})"
            self._belay_device._traceback_execute(
                src_file, src_lineno, name, cmd, record=False
            )
            # Step 2: Create the host generator that invokes ``next()`` on-device.

            def gen_inner():
                send_val = None
                try:
                    while True:
                        cmd = f"__belay_gen_next({gen_identifier}, {repr(send_val)})"
                        send_val = yield self._belay_device._traceback_execute(
                            src_file, src_lineno, name, cmd, record=False
                        )
                except StopIteration:
                    pass
                # Delete the exhausted generator on-device.
                self._belay_device(f"del {gen_identifier}")

            return gen_inner()

        executer = gen_executer if inspect.isgeneratorfunction(f) else func_executer

        if register:
            setattr(self, name, executer)

        return executer


class ThreadExecuter(Executer):
    def __call__(
        self,
        f: Optional[Callable[..., None]] = None,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = True,
    ) -> Callable[..., None]:
        """See ``Device.thread``."""
        if f is None:
            return wraps_partial(self, minify=minify, register=register, record=record)

        if self._belay_device.implementation.name == "circuitpython":
            raise FeatureUnavailableError("CircuitPython does not support threading.")

        name = f.__name__
        src_code, src_lineno, src_file = getsource(f)

        # Send the source code over to the device.
        self._belay_device(src_code, minify=minify)

        @wraps(f)
        def executer(*args, **kwargs):
            cmd = f"import _thread; _thread.start_new_thread({name}, {repr(args)}, {repr(kwargs)})"
            self._belay_device._traceback_execute(
                src_file, src_lineno, name, cmd, record=record
            )

        if register:
            setattr(self, name, executer)

        return executer
